/*----------------------------------------------------------------------
  eri.c

  High-level interface of LIBERI 

  Coded by TOYODA Masayuki, 19 Jun. 2009.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <time.h>
#include "eri_def.h"
#include "eri.h"




ERI_INLINE void coord_spherical(
  double  x,      /* (IN) Cartesian coordinate */ 
  double  y, 
  double  z,
  double *r,      /* (OUT) spherical coordinate */
  double *theta, 
  double *phi
)
{
  *r     = sqrt(x*x + y*y + z*z);
  *theta = atan2(sqrt(x*x+y*y),z);
  *phi   = atan2(y,x);
}




void ERI_Overlap(
  ERI_t                  *ptr,     /* (INOUT) ERI object */
  double                 *glF,     /* (OUT) F term */
  double                 *dglF[3], /* (OUT) derivatives of F */
  const ERI_Orbital_t *orb1,    /* (IN) orbital 1 */
  const ERI_Orbital_t *orb2,    /* (IN) orbital 2 */
  double                  cx       /* (IN) expansion center */
)
{
  int j;
  double *fk[2], *gam[2], *alp1, *alp2, *p, *F;
  double *dgam[2], *dalp[2][3], *dp[3], *dF[3];
  double ar[2], at[2], ap[2], c[3];

  int ngrid = ERI_ngrid(ptr);
  int lmax  = ERI_lmax(ptr);
  int jmax  = lmax*lmax;
  
  STEPTRACE("ERI_Overlap: in\n");

  /* expansion centers */
  c[0] = cx*orb1->c[0] + (1.0-cx)*orb2->c[0]; 
  c[1] = cx*orb1->c[1] + (1.0-cx)*orb2->c[1]; 
  c[2] = cx*orb1->c[2] + (1.0-cx)*orb2->c[2]; 
  
  coord_spherical(orb1->c[0]-c[0], orb1->c[1]-c[1], orb1->c[2]-c[2], 
                  &ar[0], &at[0], &ap[0]);
  coord_spherical(orb2->c[0]-c[0], orb2->c[1]-c[1], orb2->c[2]-c[2], 
                  &ar[1], &at[1], &ap[1]);

  /* allocate workspace memory */
  fk[0]   = (double*)malloc( ERI_Size_of_Orbital(ptr)  );
  fk[1]   = (double*)malloc( ERI_Size_of_Orbital(ptr)  );
  gam[0]  = (double*)malloc( ERI_Size_of_Gamma(ptr)    );
  gam[1]  = (double*)malloc( ERI_Size_of_Gamma(ptr)    );
  alp1    = (double*)malloc( ERI_Size_of_Alpha(ptr)    );
  alp2    = (double*)malloc( ERI_Size_of_Alpha(ptr)    );
  p       = (double*)malloc( ERI_Size_of_Overlap(ptr) );
  F       = (double*)malloc( ERI_Size_of_Overlap(ptr) );

  if (dglF != NULL) {
    dgam[0]    = (double*)malloc( ERI_Size_of_Gamma(ptr)    );
    dgam[1]    = (double*)malloc( ERI_Size_of_Gamma(ptr)    );
    dalp[0][0] = (double*)malloc( ERI_Size_of_Alpha(ptr)    );
    dalp[0][1] = (double*)malloc( ERI_Size_of_Alpha(ptr)    );
    dalp[0][2] = (double*)malloc( ERI_Size_of_Alpha(ptr)    );
    dalp[1][0] = (double*)malloc( ERI_Size_of_Alpha(ptr)    );
    dalp[1][1] = (double*)malloc( ERI_Size_of_Alpha(ptr)    );
    dalp[1][2] = (double*)malloc( ERI_Size_of_Alpha(ptr)    );
    dp[0]      = (double*)malloc( ERI_Size_of_Overlap(ptr) );
    dp[1]      = (double*)malloc( ERI_Size_of_Overlap(ptr) );
    dp[2]      = (double*)malloc( ERI_Size_of_Overlap(ptr) );
    dF[0]      = (double*)malloc( ERI_Size_of_Overlap(ptr) );
    dF[1]      = (double*)malloc( ERI_Size_of_Overlap(ptr) );
    dF[2]      = (double*)malloc( ERI_Size_of_Overlap(ptr) );
  }

  ERI_Transform_Orbital(ptr, fk[0], orb1->fr, orb1->xr, orb1->ngrid, orb1->l);
  ERI_Transform_Orbital(ptr, fk[1], orb2->fr, orb2->xr, orb2->ngrid, orb2->l);

  if (NULL == dglF) {
    ERI_LL_Gamma_d(ptr, gam[0], NULL, fk[0], ar[0]); 
    ERI_LL_Gamma_d(ptr, gam[1], NULL, fk[1], ar[1]); 
    ERI_LL_Alpha(ptr, alp1, gam[0], at[0], ap[0], orb1->l, orb1->m);
    ERI_LL_Alpha(ptr, alp2, gam[1], at[1], ap[1], orb2->l, orb2->m);
    ERI_LL_Overlap(ptr, p, alp1, alp2);
    ERI_Transform_Overlap( ptr, F, p );
    ERI_GL_Interpolate(ptr, glF, F);
  } else {
    ERI_LL_Gamma_d(ptr, gam[0], dgam[0], fk[0], ar[0]); 
    ERI_LL_Gamma_d(ptr, gam[1], dgam[1], fk[1], ar[1]); 
    ERI_LL_Alpha_d(ptr, alp1, dalp[0], gam[0], dgam[0], ar[0], at[0], ap[0], orb1->l, orb1->m);
    ERI_LL_Alpha_d(ptr, alp2, dalp[1], gam[1], dgam[1], ar[1], at[1], ap[1], orb2->l, orb2->m);
    ERI_LL_Overlap_d(ptr, p, dp, alp1, dalp[0], alp2, dalp[1], cx);
    ERI_Transform_Overlap( ptr, F, p );
    ERI_Transform_Overlap( ptr, dF[0], dp[0] );
    ERI_Transform_Overlap( ptr, dF[1], dp[1] );
    ERI_Transform_Overlap( ptr, dF[2], dp[2] );
    ERI_GL_Interpolate(ptr, glF, F);
    ERI_GL_Interpolate(ptr, dglF[0], dF[0]);
    ERI_GL_Interpolate(ptr, dglF[1], dF[1]);
    ERI_GL_Interpolate(ptr, dglF[2], dF[2]);
  }

  /* release workspace memory */
  free(fk[0]);
  free(fk[1]); 
  free(gam[0]);
  free(gam[1]); 
  free(alp1);
  free(alp2);  
  free(p);
  free(F); 

  if (NULL != dglF) {
    free(dgam[0]); 
    free(dgam[1]); 
    free(dalp[0][0]); 
    free(dalp[0][1]); 
    free(dalp[0][2]); 
    free(dalp[1][0]); 
    free(dalp[1][1]); 
    free(dalp[1][2]); 
    free(dp[0]); 
    free(dp[1]); 
    free(dp[2]); 
    free(dF[0]); 
    free(dF[1]); 
    free(dF[2]); 
  }
  
  STEPTRACE("ERI_Overlap: out\n");
}



void ERI_Integral(
  ERI_t               *ptr,
  double              I4[2], /* (OUT) integral */
  double              *dI4,   /* (OUT) derivatives */
  const ERI_Orbital_t *orb1, 
  const ERI_Orbital_t *orb2, 
  const ERI_Orbital_t *orb3, 
  const ERI_Orbital_t *orb4,
  double               scr
)
{
  double *glF, *glG, *dglF[3], *dglG[3];
  double cx1, cx2, R[3], a1[3], a2[3], a3[3], a4[3];
  clock_t clk1, clk2, clk3;
 
  int lmax_gl = ERI_lmax_gl(ptr);

  STEPTRACE("ERI_Integral: in");

  /* allocate workspace memory */
  glF = (double*)malloc( ERI_Size_of_GLF(ptr) );
  glG = (double*)malloc( ERI_Size_of_GLF(ptr) );
  if (dI4) {
    dglF[0] = (double*)malloc( ERI_Size_of_GLF(ptr) );
    dglF[1] = (double*)malloc( ERI_Size_of_GLF(ptr) );
    dglF[2] = (double*)malloc( ERI_Size_of_GLF(ptr) );
    dglG[0] = (double*)malloc( ERI_Size_of_GLF(ptr) );
    dglG[1] = (double*)malloc( ERI_Size_of_GLF(ptr) );
    dglG[2] = (double*)malloc( ERI_Size_of_GLF(ptr) );
  }

  /* expansion center */
  //cx1 = ERI_Center_DK(ptr, orb1, orb2);
  //cx2 = ERI_Center_DK(ptr, orb3, orb4);
  cx1 = ERI_Center_r2(orb1, orb2);
  cx2 = ERI_Center_r2(orb3, orb4);

  /* R */
  ERI_Coordinate_Transform(R, a1, a2, a3, a4, 
                           orb1->c, orb2->c, orb3->c, orb4->c, cx1, cx2);


  if (NULL==dI4) {
    ERI_Overlap(ptr, glF, NULL, orb1, orb2, cx1); 
    ERI_Overlap(ptr, glG, NULL, orb3, orb4, cx2);
    ERI_Integral_GL(ptr, I4, glF, glG, R[0], R[1], R[2], scr, lmax_gl);
  } else {
    ERI_Overlap(ptr, glF, dglF, orb1, orb2, cx1); 
    ERI_Overlap(ptr, glG, dglG, orb3, orb4, cx2);
    ERI_Integral_GL_d(ptr, I4, dI4, glF, glG, dglF, dglG,
                      R[0], R[1], R[2], cx1, cx2, 1e-10, scr, lmax_gl);
  }
 
  /* release workspace memory */
  free(glF);
  free(glG);
  if (dI4) {
    free(dglF[0]); 
    free(dglF[1]); 
    free(dglF[2]);
    free(dglG[0]); 
    free(dglG[1]); 
    free(dglG[2]);
  }
  
  STEPTRACE("ERI_Integral: out");
}



  
/*----------------------------------------------------------------------
  orbital_T

  kinetic energy
----------------------------------------------------------------------*/
static double orbital_T(
  ERI_t        *ptr,
  const double *fr,     
  const double *xr,     
  int           nrgrid, 
  int           l       
)
{
  int i;
  double k, af, ke, dk;
  double *fk;

  int           nkgrid = ERI_ngrid(ptr);
  const double *kmesh  = ERI_Mesh_Array_k(ptr);
 
  fk = (double*)malloc( ERI_Size_of_Orbital(ptr) );
  ERI_Transform_Orbital(ptr, fk, fr, xr, nrgrid, l);

  ke = 0.0;
  for (i=0; i<nkgrid; i++) {
    k  = ERI_Mesh_k(ptr, i);
    dk = ERI_Mesh_dk(ptr, i);
    af = fk[2*i+0]*fk[2*i+0]+fk[2*i+1]*fk[2*i+1];
    ke += 8.0*(k*k*k*k)*af*dk;
  }
  
  free(fk);

  return ke;
}




/*----------------------------------------------------------------------
  orbital_r2

  <r^2>
----------------------------------------------------------------------*/
static double orbital_r2(
  const double *fr,
  const double *xr,
  int           ngrid
)
{
  int ir;
  double norm, br2k, x0, x1, r2, r, f;

  norm = 0.0;
  br2k = 0.0;
  for (ir=0; ir<ngrid-1; ir++) {
    x0 = xr[ir];
    x1 = xr[ir+1];
    r  = 0.5*(x0+x1);
    r2 = r*r;
    f  = 0.5*(fr[ir]+fr[ir+1]);
    norm += r2*f*f*(x1-x0);
    br2k += r2*r2*f*f*(x1-x0);
  }

  return br2k/norm;
}




/*----------------------------------------------------------------------
  orbital_r

  <r>
----------------------------------------------------------------------*/
static double orbital_r(
  const double *fr,
  const double *xr,
  int           ngrid
)
{
  int ir;
  double norm, brk, x0, x1, r, r2, f;

  norm = 0.0;
  brk  = 0.0;
  for (ir=0; ir<ngrid-1; ir++) {
    x0 = xr[ir];
    x1 = xr[ir+1];
    r  = 0.5*(x0+x1);
    r2 = r*r;
    f  = 0.5*(fr[ir]+fr[ir+1]);
    norm += r2*f*f*(x1-x0);
    brk  += r2*r*f*f*(x1-x0);
  }

  return brk/norm;
}




void ERI_Coordinate_Transform(
  double out_R[3],  /* (OUT) displacement in spherical coord. */
  double out_a1[3], /* (OUT) translated center 1 in spherical coord. */
  double out_a2[3], /* (OUT) translated center 2 in spherical coord. */
  double out_a3[3], /* (OUT) translated center 3 in spherical coord. */
  double out_a4[3], /* (OUT) translated center 4 in spherical coord. */
  const double in_a1[3],  /* (IN) center 1 in Cartesian coord. */
  const double in_a2[3],  /* (IN) center 2 in Cartesian coord. */
  const double in_a3[3],  /* (IN) center 3 in Cartesian coord. */
  const double in_a4[3],  /* (IN) center 4 in Cartesian coord. */
  double x12,   /* (IN) ratio */
  double x34    /* (IN) ratio */
)
{
  double c12[3], c34[3];

  c12[0] = x12*in_a1[0] + (1.0-x12)*in_a2[0];
  c12[1] = x12*in_a1[1] + (1.0-x12)*in_a2[1];
  c12[2] = x12*in_a1[2] + (1.0-x12)*in_a2[2];
  c34[0] = x34*in_a3[0] + (1.0-x34)*in_a4[0];
  c34[1] = x34*in_a3[1] + (1.0-x34)*in_a4[1];
  c34[2] = x34*in_a3[2] + (1.0-x34)*in_a4[2];

  coord_spherical(in_a1[0]-c12[0], in_a1[1]-c12[1], in_a1[2]-c12[2], 
                  &out_a1[0], &out_a1[1], &out_a1[2]);
  coord_spherical(in_a2[0]-c12[0], in_a2[1]-c12[1], in_a2[2]-c12[2],
                  &out_a2[0], &out_a2[1], &out_a2[2]);
  coord_spherical(in_a3[0]-c34[0], in_a3[1]-c34[1], in_a3[2]-c34[2], 
                  &out_a3[0], &out_a3[1], &out_a3[2]);
  coord_spherical(in_a4[0]-c34[0], in_a4[1]-c34[1], in_a4[2]-c34[2], 
                  &out_a4[0], &out_a4[1], &out_a4[2]);
 
  coord_spherical(c34[0]-c12[0], c34[1]-c12[1], c34[2]-c12[2], 
                  &out_R[0], &out_R[1], &out_R[2]);
}

 


/*----------------------------------------------------------------------
  ERI_Center_r2
----------------------------------------------------------------------*/
double ERI_Center_r2(
  const ERI_Orbital_t *pA,
  const ERI_Orbital_t *pB
)
{
   double d1 = orbital_r2(pA->fr, pA->xr, pA->ngrid);
   double d2 = orbital_r2(pB->fr, pB->xr, pB->ngrid);

   return d2/(d1+d2);
}




/*----------------------------------------------------------------------
  ERI_Center_DK
----------------------------------------------------------------------*/
double ERI_Center_DK(
  ERI_t        *ptr,
  const ERI_Orbital_t *pA,
  const ERI_Orbital_t *pB
)
{
  double d1 = orbital_r2(pA->fr, pA->xr, pA->ngrid);
  double d2 = orbital_r2(pB->fr, pB->xr, pB->ngrid);
  double k1 = orbital_T(ptr, pA->fr, pA->xr, pA->ngrid, pA->l);
  double k2 = orbital_T(ptr, pB->fr, pB->xr, pB->ngrid, pB->l);
   
  return (d2*k1)/(d2*k1+d1*k2);
}


/* EOF */
