/*----------------------------------------------------------------------
  eri_gaunt_table.c

  Table of the Gaunt coefficients.

  16 Feb. 2009, M. Toyoda
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "eri_sf.h"
#include "eri_gaunt_table.h"


#define THRESHOLD  1e-10


/*----------------------------------------------------------------------
  INTERNAL SUBROUTINES
----------------------------------------------------------------------*/
typedef struct {
  double coeff;
  int    j1;
  int    j2;
} item;


typedef struct {
  item *items;
  int   n;
} table;



struct ERI_Gaunt_Table_struct {
  int lmax;
  table *table;
};



/*----------------------------------------------------------------------
  ERI_GTable_Init
----------------------------------------------------------------------*/
ERI_Gaunt_Table_t* ERI_Gaunt_Table_Init(int lmax)
{
  int i, n, j, l, m, l1, m1, l2, m2, lsum;
  double g;
  item *buffer;
  ERI_Gaunt_Table_t *ptr;

  int jmax = lmax*lmax;
  int nmax = jmax*jmax;

  ptr = (ERI_Gaunt_Table_t*)malloc(sizeof(struct ERI_Gaunt_Table_struct));
  if (NULL==ptr) { return NULL; }
 
  ptr->table =  (table*)malloc(sizeof(table)*jmax);
  if (NULL==ptr->table) { 
    ERI_Gaunt_Table_Free(ptr);
    return NULL;
  }

  buffer = (item*)malloc(sizeof(item)*nmax);
  if (NULL==buffer) {
    ERI_Gaunt_Table_Free(ptr);
    return NULL;
  }

  for (j=0; j<jmax; j++) {
    l = (int)sqrt((double)j);
    m = j-l*(l+1);
    n=0;
    for (l1=0; l1<lmax; l1++) {
      for (l2=0; l2<lmax; l2++) {
        lsum = l+l1+l2; 
        if (l<=l1+l2 && l1<=l+l2 && l2<=l+l1 && lsum%2==0) {
          for (m1=-l1; m1<=l1; m1++) {
            m2 = m-m1;
            if (-l2<=m2 && m2<=l2) {
              g = ERI_Gaunt(l, m, l1, m1, l2, m2);
              if (fabs(g)>=THRESHOLD) {
                buffer[n].coeff =  g;
                buffer[n].j1 = l1*(l1+1)+m1;
                buffer[n].j2 = l2*(l2+1)+m2;
                n++;
              }
            }
          } /* end loop of m1 */
        }
      } /* end loop of l2 */
    } /* end loop of l1 */
    ptr->table[j].n = n;
    ptr->table[j].items = (item*)malloc(sizeof(item)*n);
    for (i=0; i<n; i++) {
      ptr->table[j].items[i].coeff = buffer[i].coeff;
      ptr->table[j].items[i].j1 = buffer[i].j1;
      ptr->table[j].items[i].j2 = buffer[i].j2;
    }
  }
  ptr->lmax = lmax;

  return ptr;
}



/*----------------------------------------------------------------------
  ERI_Gaunt_Table_Required_Size
----------------------------------------------------------------------*/
size_t ERI_Gaunt_Table_Required_Size(int lmax)
{
  int j, jmax, n;
  int l, m, l1, m1, l2, m2, lsum;
  double g;
  size_t memsize;

  jmax = lmax*lmax;
  memsize = sizeof(table)*jmax;
  
  for (j=0; j<jmax; j++) {
    l = (int)sqrt((double)j);
    m = j-l*(l+1);
    n=0;
    for (l1=0; l1<lmax; l1++) {
      for (l2=0; l2<lmax; l2++) {
        lsum = l+l1+l2; 
        if (l<=l1+l2 && l1<=l+l2 && l2<=l+l1 && lsum%2==0) {
          for (m1=-l1; m1<=l1; m1++) {
            m2 = m-m1;
            if (-l2<=m2 && m2<=l2) {
              g = ERI_Gaunt(l, m, l1, m1, l2, m2);
              if (fabs(g)>=THRESHOLD) { n++; }
            }
          } /* end loop of m1 */
        }
      } /* end loop of l2 */
    } /* end loop of l1 */
    memsize += sizeof(item)*n; /* my_table[j].items */
  } /* end loop of j */
  
  return memsize;
}




/*----------------------------------------------------------------------
  GTable_Free()
----------------------------------------------------------------------*/
void ERI_Gaunt_Table_Free(ERI_Gaunt_Table_t *ptr)
{
  int j, jmax;

  if (ptr) {
    if (ptr->table) {
      jmax = ptr->lmax*ptr->lmax; 
      for (j=0; j<jmax; j++) { free(ptr->table[j].items); }
      free(ptr->table);
    }
    free(ptr);
  }
}




/*----------------------------------------------------------------------
  ERI_GTable_Num_Items
 
  This returns the number of non-zero Gaunt coefficients of which the 
  first angular momentum equals to j. 
----------------------------------------------------------------------*/
int ERI_Gaunt_Table_Num_Items(const ERI_Gaunt_Table_t* ptr, int j) 
{
  return ptr->table[j].n; 
} 




/*----------------------------------------------------------------------
  ERI_GTable_Coeff
 
  This returns index-th Gaunt coefficient of which the first angular 
  momentum equalrs to j.

  index must satisfy 0 <= index <= ERI_GauntTable_Num_Items(j).
----------------------------------------------------------------------*/
double ERI_Gaunt_Table_Coeff(const ERI_Gaunt_Table_t* ptr, int j, int index)
{
  return ptr->table[j].items[index].coeff;
}
  


/*----------------------------------------------------------------------
  ERI_Gaunt_Table_j1j2
  
  This returns index-th (j1,j2) pair of which the first angular 
  momentum equals to j.
----------------------------------------------------------------------*/
void ERI_Gaunt_Table_j1j2(
  const ERI_Gaunt_Table_t* ptr,
  int j,     /* (IN) the first angular momentum */
  int index, /* (IN) index */
  int *j1,   /* (OUT) the second angular momentum */
  int *j2    /* (OUT) the third angular momentum */
)
{
  *j1 = ptr->table[j].items[index].j1;
  *j2 = ptr->table[j].items[index].j2;
}


/* EOF */
