/*----------------------------------------------------------------------
  eri_ll.c

  Low-level functions of LIBERI

  Coded by TOYODA Masayuki, June 19, 2009
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "eri_def.h"
#include "eri.h"
#include "eri_sf.h"
#include "eri_interpolate.h"
#include "eri_gaunt_table.h"
#include "sbt/linear/eri_linfsbt.h"
#include "sbt/log/eri_logfsbt.h"


/*--- Switches for speeding-up ---*/
#define GAUNTTABLE 1
#define SUMSKIP    1
#define SUMSKIP_THRESHOLD 1e-10


static char* my_version = "ALPHA-090619";


struct ERI_Struct {
  int lmax;
  int lmax_gl;

  /* workspace */
  double *sb;
  double *dsb;
  double *in;
  double *out;

  /* index table */
  int    *j2l;
  int    *j2m;

  /* Gauss-Laguerre quadrature */
  int glq_n;
  double *glq_x;
  double *glq_w;
  double *glq_j;
  double *glq_dj;

  /* SBT */
  int sbttype;
  ERI_SBT_t  *sbt;

  /* Gaunt-coefficients */
  ERI_Gaunt_Table_t *gtbl;

  ERI_Init_Misc info;
};
 



/*----------------------------------------------------------------------
  INTERNAL SUBROUTINES
----------------------------------------------------------------------*/
ERI_INLINE void phase(int l, double *a)
{
  double re = a[0], im = a[1];

  switch ((l%4+4)%4) {
  case 0: break;
  case 1: a[0] = -im; a[1] =  re; break;
  case 2: a[0] = -re; a[1] = -im; break;
  case 3: a[0] =  im; a[1] = -re; break;
  }
}



const char* ERI_Version(void)
{
  return my_version;
}




/*----------------------------------------------------------------------
  ERI_Init
----------------------------------------------------------------------*/
ERI_t* ERI_Init(
  int                  lmax,
  int                  lmax_gl,
  int                  ngrid,
  int                  ngl,
  const ERI_Init_Misc *info
)
{
  int j, l, m;
  ERI_t *ptr;
    
  double dt;
  
  int jmax  = lmax*lmax;

  STEPTRACE("ERI_Init: in");
 
  /* allocation */ 
  ptr = (ERI_t*)malloc(sizeof(ERI_t));
  if (NULL==ptr) { return NULL; }

  ptr->sb     = (double*)malloc(sizeof(double)*(lmax+2)*ngrid);
  ptr->dsb    = (double*)malloc(sizeof(double)*(lmax+2)*ngrid);
  ptr->in     = (double*)malloc(sizeof(double)*ngrid*2);
  ptr->out    = (double*)malloc(sizeof(double)*ngrid*2);
  ptr->j2l    = (int*)malloc(sizeof(int)*jmax);
  ptr->j2m    = (int*)malloc(sizeof(int)*jmax);
  ptr->glq_x  = (double*)malloc(sizeof(double)*ngl); 
  ptr->glq_w  = (double*)malloc(sizeof(double)*ngl); 
  ptr->glq_j  = (double*)malloc(sizeof(double)*ngl*lmax); 
  ptr->glq_dj = (double*)malloc(sizeof(double)*ngl*lmax); 
  if (NULL==ptr->sb || NULL==ptr->dsb || NULL==ptr->in 
    || NULL==ptr->out || NULL==ptr->j2l || NULL==ptr->j2m
    || NULL==ptr->glq_x || NULL==ptr->glq_w || NULL==ptr->glq_j 
    || NULL==ptr->glq_dj ) {
    ERI_Free(ptr);
    return NULL;
  }
  
  /* j <-> (l,m) conversion table */
  for (j=0; j<jmax; j++) {
    l = (int)sqrt((double)j);
    m = j-l*(l+1);
    ptr->j2l[j] = l;
    ptr->j2m[j] = m;
  }
  
  /* abscissas and weights for Gauss-Laguerre integration */
  ERI_GLQ(ptr->glq_x, ptr->glq_w, ngl);

  /* Gaunt coefficients */
  ptr->gtbl = ERI_Gaunt_Table_Init(lmax);

  /* additional parameters */
  if (NULL==info) {
    /* default values */
    ptr->info.sbttype = ERI_SBT_LINEAR;
    ptr->info.rmax    = 100.0;
    ptr->info.rho0    = -10.0;
    ptr->info.qmin    = 1e-4;
    ptr->info.qmax    = 1e-2;
    ptr->info.nq      = 3; 
  } else {
    ptr->info.sbttype = info->sbttype;
    ptr->info.rmax    = info->rmax;
    ptr->info.rho0    = info->rho0;
    ptr->info.qmin    = info->qmin;
    ptr->info.qmax    = info->qmax;
    ptr->info.nq      = info->nq; 
  }
  dt = 2.0*PI/(log(ptr->info.rmax)-ptr->info.rho0);

  switch (ptr->info.sbttype) {
  case ERI_SBT_LINEAR:
    ptr->sbt = ERI_LinFSBT_Init(ngrid, lmax, ptr->info.rmax);
    break;
  case ERI_SBT_LOG:
    ptr->sbt = ERI_LogFSBT_Init(lmax, ngrid, ptr->info.rho0, dt, 
                                ptr->info.qmin, ptr->info.qmax, ptr->info.nq);

    break;
  }
  if (NULL==ptr->sbt) {
    ERI_Free(ptr);
    return NULL;
  }

  ptr->lmax    = lmax;
  ptr->lmax_gl = lmax_gl;
  ptr->glq_n   = ngl;
  //ptr->sbttype = sbttype;

  return ptr;
}




/*----------------------------------------------------------------------
  ERI_Required_Size
----------------------------------------------------------------------*/
size_t ERI_Required_Size(
  int                  lmax,
  int                  lmax_gl,
  int                  ngrid,
  int                  ngl,
  const ERI_Init_Misc *info
)
{
  int nq, type;
  size_t sz;

  int jmax = lmax*lmax;
 
  sz = sizeof(ERI_t)
       + ERI_Gaunt_Table_Required_Size(lmax)
       + sizeof(double)*(lmax+2)*ngrid /* sb */
       + sizeof(double)*(lmax+2)*ngrid /* dsb */
       + sizeof(double)*ngrid*2        /* in */
       + sizeof(double)*ngrid*2        /* out */
       + sizeof(int)*jmax              /* j2l */
       + sizeof(int)*jmax              /* j2m */
       + sizeof(double)*ngl            /* glx */
       + sizeof(double)*ngl            /* glw */
       + sizeof(double)*ngl*lmax       /* glq_j */
       + sizeof(double)*ngl*lmax       /* glq_dj */
    ;

  type = info ? info->sbttype : ERI_SBT_LINEAR;

  switch (type) {
  case ERI_SBT_LOG:
    if (NULL==info) {
      nq = 3;
    } else {
      nq = info->nq; 
    }
    sz += ERI_LogFSBT_Required_Size(lmax, ngrid, nq);
    break;
  case ERI_SBT_LINEAR:
    sz += ERI_LinFSBT_Required_Size(ngrid, lmax);
    break;
  default:
    abort();
  }

  return sz;
}




/*----------------------------------------------------------------------
  ERI_Free
----------------------------------------------------------------------*/
void ERI_Free(ERI_t *ptr)
{
  if (ptr) {
    if (ptr->sb)    { free(ptr->sb);    }
    if (ptr->dsb)   { free(ptr->dsb);   }
    if (ptr->in)    { free(ptr->in);    }
    if (ptr->out)   { free(ptr->out);   }
    if (ptr->j2l)   { free(ptr->j2l);   }
    if (ptr->j2m)   { free(ptr->j2m);   }
    if (ptr->glq_x) { free(ptr->glq_x); }
    if (ptr->glq_w) { free(ptr->glq_w); }
    if (ptr->glq_j) { free(ptr->glq_j); }
    if (ptr->sbt)   { ERI_SBT_Free(ptr->sbt); }
    if (ptr->gtbl)  { ERI_Gaunt_Table_Free(ptr->gtbl); }
    free(ptr); 
  }
}



int ERI_lmax(const ERI_t *ptr)  { return ptr->lmax; }
int ERI_lmax_gl(const ERI_t *ptr) { return ptr->lmax_gl; }
int ERI_ngl(const ERI_t *ptr)   { return ptr->glq_n; }

const double* ERI_Mesh_Array_glx(const ERI_t *ptr) { return ptr->glq_x; }
const double* ERI_Mesh_Array_glw(const ERI_t *ptr) { return ptr->glq_w; }

int ERI_ngrid(const ERI_t *ptr) 
{ 
  return ERI_SBT_ngrid(ptr->sbt);
}

double ERI_Mesh_r(const ERI_t *ptr, int i) 
{ 
  return ERI_SBT_Mesh_r(ptr->sbt, i);
}


double ERI_Mesh_k(const ERI_t *ptr, int i)
{ 
  return ERI_SBT_Mesh_k(ptr->sbt, i);
}


const double* ERI_Mesh_Array_r(const ERI_t *ptr)
{ 
  return ERI_SBT_Mesh_Array_r(ptr->sbt);
}


const double* ERI_Mesh_Array_k(const ERI_t *ptr)
{ 
  return ERI_SBT_Mesh_Array_k(ptr->sbt);
}


double ERI_Mesh_dr(const ERI_t *ptr, int i) 
{ 
  return ERI_SBT_Mesh_dr(ptr->sbt, i);
}


double ERI_Mesh_dk(const ERI_t *ptr, int i)
{ 
  return ERI_SBT_Mesh_dk(ptr->sbt, i);
}



/*----------------------------------------------------------------------
  SIZE Calculators  
----------------------------------------------------------------------*/
size_t ERI_Size_of_Orbital(const ERI_t *ptr) 
{
  int ngrid = ERI_ngrid(ptr);
  return sizeof(double)*ngrid*2;
}


size_t ERI_Size_of_Gamma(const ERI_t *ptr)
{
  int ngrid = ERI_ngrid(ptr);
  int lmax  = ERI_lmax(ptr);
  int jmax  = lmax*lmax;

#if SUMSKIP
  return sizeof(double)*(ngrid*jmax*2 + jmax);
#else
  return sizeof(double)*(ngrid*jmax*2);
#endif
}


size_t ERI_Size_of_Alpha(const ERI_t *ptr)
{
  int ngrid = ERI_ngrid(ptr);
  int lmax  = ERI_lmax(ptr);
  int jmax  = lmax*lmax;

#if SUMSKIP
  return sizeof(double)*(ngrid*jmax*2 + jmax);
#else
  return sizeof(double)*(ngrid*jmax*2);
#endif
}


size_t ERI_Size_of_Overlap(const ERI_t *ptr)
{
  int ngrid = ERI_ngrid(ptr);
  int lmax  = ERI_lmax(ptr);
  int jmax  = lmax*lmax;

#if SUMSKIP
  return sizeof(double)*(ngrid*jmax*2 + jmax); 
#else
  return sizeof(double)*(ngrid*jmax*2); 
#endif
}


size_t ERI_Size_of_GLF(const ERI_t *ptr)
{
  int ngl   = ERI_ngl(ptr);
  int lmax0 = ERI_lmax_gl(ptr);
  int jmax0 = lmax0*lmax0;

#if SUMSKIP 
  return sizeof(double)*(ngl*jmax0*2 + jmax0); 
#else
  return sizeof(double)*(ngl*jmax0*2); 
#endif
}




/*----------------------------------------------------------------------
  ERI_Orbital

  For a given orbital function, this performs sph. Bessel transform.
----------------------------------------------------------------------*/
void ERI_Transform_Orbital(
  ERI_t        *ptr,
  double       *fk, /* (OUT) transformed orbital */
  const double *fr, /* (IN)  radial orbital function */
  const double *xr, /* (IN)  radial grid for fr */
  int           n,  /* (IN)  number of radial grid points */
  int           l   /* (IN)  angular momentum */
)
{
  int i;
  ERI_CSpline_t *cs;

  int           ngrid = ERI_ngrid(ptr);
  const double *rmesh = ERI_Mesh_Array_r(ptr);
  double       *in    = ptr->in; 
  ERI_SBT_t    *sbt   = ptr->sbt;

  /* change radial grid */ 
  {
    cs = ERI_CSpline_Init(xr, fr, n);
    for (i=0; i<ngrid; i++) {
      in[2*i+0] = ERI_CSpline_Eval(rmesh[i], cs);
      in[2*i+1] = 0.0;
    }
    ERI_CSpline_Free(cs);
  }
  
  /* transform */
  ERI_SBT_Transform(sbt, fk, in, l, ERI_SBT_FORWARD);
}




/*----------------------------------------------------------------------
  ERI_Gamma
  g  : Gamma factor (out)
  dg : Deribative of Gamma factor (out)
  fk : Transformed wave function 
  R  : Displacement
----------------------------------------------------------------------*/
void ERI_LL_Gamma_d(
  ERI_t        *ptr,
  double       *g,
  double       *dg,
  const double *fk,
  double        rd
) 
{  
  int l1, l2, ik, ig, igbase;
  int j;
#if SUMSKIP
  double *gmax  = NULL;
  double *dgmax = NULL;
#endif

  int           ngrid = ERI_ngrid(ptr);
  int           lmax  = ERI_lmax(ptr);
  int           jmax  = lmax*lmax;
  const double *kmesh = ERI_Mesh_Array_k(ptr);
  double       *sb    = ptr->sb;
  double       *dsb   = ptr->dsb;
  double       *in    = ptr->in;
  double       *out   = ptr->out;
  ERI_SBT_t    *sbt   = ptr->sbt;

#if SUMSKIP
  gmax  =  &g[2*ngrid*jmax];
  for (j=0; j<jmax; j++) { gmax[j] = 0.0; }
  if (dg) {
    dgmax =  &dg[2*ngrid*jmax];
    for (j=0; j<jmax; j++) { dgmax[j] = 0.0; }
  } 
#endif

  /*--- the spherical Bessel functions ---*/  
  for (ik=0; ik<ngrid; ik++) {
    ERI_Spherical_Bessel(kmesh[ik]*rd, lmax, &sb[ik*lmax], &dsb[ik*lmax]);
  }

  for (l2=0; l2<lmax; l2++) {
    for (ik=0; ik<ngrid; ik++) {
      in[2*ik+0] = fk[2*ik+0]*sb[ik*lmax+l2];
      in[2*ik+1] = fk[2*ik+1]*sb[ik*lmax+l2];
    }
    ERI_SBT_Transform_Input(sbt, in, ERI_SBT_BACKWARD);
    for (l1=0; l1<lmax; l1++) {
      j = l1*lmax+l2;
      igbase = j*ngrid;
      ERI_SBT_Transform_Output(sbt, out, l1);
      for (ik=0; ik<ngrid; ik++) {
        ig = igbase+ik;
        g[2*ig+0] = out[2*ik+0]*2.0/PI;
        g[2*ig+1] = out[2*ik+1]*2.0/PI;
#if SUMSKIP
        if (fabs(g[2*ig+0])>gmax[j]) gmax[j] = fabs(g[2*ig+0]);
        if (fabs(g[2*ig+1])>gmax[j]) gmax[j] = fabs(g[2*ig+1]);
#endif
      } /* loop of ik */
    } /* loop of l1 */
  } /* loop of l2 */

  if (dg) {
    /*--- deribative ---*/
    for (l2=0; l2<lmax; l2++) {
      for (ik=0; ik<ngrid; ik++) {
        in[2*ik+0] = fk[2*ik+0]*dsb[ik*lmax+l2]*kmesh[ik];
        in[2*ik+1] = fk[2*ik+1]*dsb[ik*lmax+l2]*kmesh[ik];
      }
      ERI_SBT_Transform_Input(sbt, in, ERI_SBT_BACKWARD);
      for (l1=0; l1<lmax; l1++) {
        j = l1*lmax+l2;
        igbase = j*ngrid;
        ERI_SBT_Transform_Output(sbt, out, l1);
        for (ik=0; ik<ngrid; ik++) {
          dg[2*(igbase+ik)+0] = out[2*ik+0]*2.0/PI;
          dg[2*(igbase+ik)+1] = out[2*ik+1]*2.0/PI;
#if SUMSKIP
          if (fabs(dg[2*(igbase+ik)+0])>dgmax[j]) dgmax[j] = fabs(dg[2*(igbase+ik)+0]);
          if (fabs(dg[2*(igbase+ik)+1])>dgmax[j]) dgmax[j] = fabs(dg[2*(igbase+ik)+1]);
#endif
        } /* loop of ik */
      } /* loop of l1 */
    } /* loop of l2 */
  } /* end if */
}





/*----------------------------------------------------------------------
  Alpha term 
 
  Alpha term.
  If you need the ferivatives as well, you should call ERI_Alpha_d.

  This routine uses the complex spherical harmonic functions.

  alpha[l1][m2](r) = 4 pi sum[l2][m2] i^(l-l1+l2) G(l,m,l1,m1,l2,m2) 
                                        Y[l2][m2](anlge R) g[l1][l2](r)
----------------------------------------------------------------------*/
void ERI_LL_Alpha(
  ERI_t        *ptr,
  double       *alp,   /* (OUT) Alpha term*/
  const double *g,     /* (IN) Gamma term */ 
  double        theta, /*      in the spherical coordinates. */
  double        phi,      
  int           l,     /* (IN) Angular momentum of the orbital */
  int           m
)
{
  int j, j1, j2, ir;
  int ia, ia_base, ig, ig_base;
  int l1, m1, l2, m2;
  double sh[2], dsht[2], dshp[2], gnt, fpg;
#if GAUNTTABLE
  int i, n;
#endif
#if SUMSKIP
  double *amax = NULL;
  const double *gmax = NULL;
#endif
 
  int        ngrid = ERI_ngrid(ptr);
  int        lmax  = ERI_lmax(ptr);
  int        jmax  = lmax*lmax;
  const int *j2l   = ptr->j2l;
  const int *j2m   = ptr->j2m;
  const ERI_Gaunt_Table_t * gtbl = ptr->gtbl;

  STEPTRACE("ERI_Alpha: in ");

#if SUMSKIP
  amax  = &alp[2*ngrid*jmax];
  gmax  = &g[2*ngrid*jmax];
  for (j1=0; j1<jmax; j1++) { amax[j1] = 0.0; }
#endif

  j = l*(l+1)+m;

  /* zero reset */
  for (ia=0; ia<ngrid*jmax*2; ia++) { alp[ia] = 0.0; }
 
#if GAUNTTABLE
  n = ERI_Gaunt_Table_Num_Items(gtbl, j);
  for (i=0; i<n; i++) {
    gnt = ERI_Gaunt_Table_Coeff(gtbl, j, i);
    ERI_Gaunt_Table_j1j2(gtbl, j, i, &j1, &j2);
    fpg = gnt*4.0*PI;
    l1 = j2l[j1];
    m1 = j2m[j1];
    l2 = j2l[j2];
    m2 = j2m[j2];
#if SUMSKIP
    if (gmax[l1*lmax+l2]>SUMSKIP_THRESHOLD) 
#endif
    {
      ERI_Spherical_Harmonics(l2, m2, theta, phi, sh, dsht, dshp);
      phase(l-l1+l2, sh);
      sh[0] = fpg*sh[0];
      sh[1] = fpg*sh[1];
      ia_base = j1*ngrid;
      ig_base = (l1*lmax+l2)*ngrid;
      for (ir=0; ir<ngrid; ir++) {
        alp[2*(ia_base+ir)+0] 
          += sh[0]*g[2*(ig_base+ir)+0] - sh[1]*g[2*(ig_base+ir)+1]; 
        alp[2*(ia_base+ir)+1] 
          += sh[1]*g[2*(ig_base+ir)+0] + sh[0]*g[2*(ig_base+ir)+1];
#if 0
        ia = ia_base+ir;
        ig = ig_base+ir;
        alp[2*ia+0] += sh[0]*g[2*ig+0] - sh[1]*g[2*ig+1]; 
        alp[2*ia+1] += sh[1]*g[2*ig+0] + sh[0]*g[2*ig+1];
#endif
      } 
    }
  } /* loop of i */
#if SUMSKIP
  for (j1=0; j1<jmax; j1++) {
    ia_base = j1*ngrid;
    for (ir=0; ir<ngrid; ir++) {
      ia = (ia_base+ir)*2;
      if (fabs(alp[ia+0])>amax[j1]) amax[j1] = fabs(alp[ia+0]);
      if (fabs(alp[ia+1])>amax[j1]) amax[j1] = fabs(alp[ia+1]);
    } 
  }
#endif
#else /* GAUNTTABLE */
  for (j1=0; j1<jmax; j1++) {
    l1 = j2l[j1];
    m1 = j2m[j1];
    ia_base = j1*ngrid;
    for (j2=0; j2<jmax; j2++) {
      l2 = j2l[j2];
      m2 = j2m[j2];
#if SUMSKIP
      if (gmax[l1*lmax+l2]>SUMSKIP_THRESHOLD) {
#endif
        gnt = ERI_Gaunt(l, m, l1, m1, l2, m2);
        fpg = gnt*4.0*PI;
        ERI_Spherical_Harmonics(l2, m2, theta, phi, sh, dsht, dshp);
        phase(l-l1+l2, sh);
        sh[0] = fpg*sh[0];
        sh[1] = fpg*sh[1];
        ig_base = (l1*lmax+l2)*ngrid;
        for (ir=0; ir<ngrid; ir++) {
          ia = (ia_base+ir)*2;
          ig = (ig_base+ir)*2;
          alp[ia+0] += sh[0]*g[ig+0] - sh[1]*g[ig+1]; 
          alp[ia+1] += sh[1]*g[ig+0] + sh[0]*g[ig+1];
        }
#if SUMSKIP
      }
#endif 
    }  /* loop of j2 */
#if SUMSKIP
    for (ir=0; ir<ngrid; ir++) {
      ia = (ia_base+ir)*2;
      if (fabs(alp[ia+0])>amax[j1]) amax[j1] = fabs(alp[ia+0]);
      if (fabs(alp[ia+1])>amax[j1]) amax[j1] = fabs(alp[ia+1]);
    }
#endif
  } /* loop of j1 */
#endif /* GAUNTTABLE */

  STEPTRACE("ERI_Alpha: out");
}




/*----------------------------------------------------------------------
  ERI_Alpha_d

  This calculates the Alpha term and its derivatives.
  If you don't need the derivatives, you may call ERI_Alpha which is 
  faster than this routine.
----------------------------------------------------------------------*/
void ERI_LL_Alpha_d(
  ERI_t        *ptr,
  double       *alp,     /* (OUT) Alpha term */
  double       *dalp[3], /* (OUT) Derivatives of Alpha term */
  const double *g,       /* (IN)  Gamma term */
  const double *dg,      /* (IN)  Derivatives of Gamma term */
  double        rd,      /* (IN)  Displacement from the expansion center */
  double        theta,   /*       in the spherical coordinates. */
  double        phi,
  int           l,       /* (IN)  Angular mementum of the orbital */
  int           m
)
{
  int i, j, ir, j1, j2;
  int ia, ia_base, ig, ig_base;
  int l1, m1, l2, m2;

  double sh[2], dsht[2], dshp[2], gnt, fpg;
  double dgsh[2], gdsht[2], gdshp[2];
 
  double ct = cos(theta);
  double st = sin(theta);
  double cp = cos(phi);
  double sp = sin(phi);
  double x1, y1, z1, x2, y2, z2, x3, y3;

#if GAUNTTABLE
  int n;
#endif

#if SUMSKIP
  double *amax, *damax[3];
  const double *gmax, *dgmax;
#endif 
  
  const double rd_thresh = 1e-10;
 
  int        ngrid = ERI_ngrid(ptr);
  int        lmax  = ERI_lmax(ptr);
  int        jmax  = lmax*lmax;
  const int *j2l   = ptr->j2l;
  const int *j2m   = ptr->j2m;
  const ERI_Gaunt_Table_t *gtbl = ptr->gtbl;

#if SUMSKIP
  amax     = &alp[2*ngrid*jmax];
  damax[0] = &dalp[0][2*ngrid*jmax];
  damax[1] = &dalp[1][2*ngrid*jmax];
  damax[2] = &dalp[2][2*ngrid*jmax];
  gmax     = &g[2*ngrid*jmax];
  dgmax    = &dg[2*ngrid*jmax];
  for (j1=0; j1<jmax; j1++) {
    amax[j1] = 0.0;
    damax[0][j1] = 0.0;
    damax[1][j1] = 0.0;
    damax[2][j1] = 0.0;
  }
#endif 
   
  /* zero reset */ 
  for (ia=0; ia<ngrid*jmax*2; ia++) {
    alp[ia] = 0.0;
    dalp[0][ia] = 0.0;
    dalp[1][ia] = 0.0;
    dalp[2][ia] = 0.0;
  }

  if (fabs(rd)<rd_thresh) {
    /* The both orbitals are on a same site,
       The derivatives are not to be calculated 
       but just set to zero. */
    ERI_LL_Alpha(ptr, alp, g, theta, phi, l, m);
  } else {
    /* The orbitals are on different sites.
       The derivatives are calculated as well. */

    j = l*(l+1)+m;
  
    x1 = st*cp;    y1 = st*sp;    z1 = ct;
    x2 = ct*cp/rd; y2 = ct*sp/rd; z2 = -st/rd;
    x3 = -sp/rd/st; y3 = cp/rd/st;

#if GAUNTTABLE
    n = ERI_Gaunt_Table_Num_Items(gtbl, j);
    for (i=0; i<n; i++) {
      gnt = ERI_Gaunt_Table_Coeff(gtbl, j, i);
      ERI_Gaunt_Table_j1j2(gtbl, j, i, &j1, &j2);
      fpg = gnt*4.0*PI;
      l1 = j2l[j1];
      m1 = j2m[j1];
      l2 = j2l[j2];
      m2 = j2m[j2];

      ERI_Spherical_Harmonics(l2, m2, theta, phi, sh, dsht, dshp);
      phase(l-l1+l2, sh);
      phase(l-l1+l2, dsht);
      phase(l-l1+l2, dshp); 
      sh[0]   = fpg*sh[0];
      sh[1]   = fpg*sh[1];
      dsht[0] = fpg*dsht[0];
      dsht[1] = fpg*dsht[1];
      dshp[0] = fpg*dshp[0];
      dshp[1] = fpg*dshp[1];
      ia_base  = j1*ngrid;
      ig_base  = (l1*lmax+l2)*ngrid;
#if SUMSKIP
      if (gmax[l1*lmax+l2]>SUMSKIP_THRESHOLD) 
#endif
      {
        for (ir=0; ir<ngrid; ir++) {
          alp[2*(ia_base+ir)+0] += sh[0]*g[2*(ig_base+ir)+0] - sh[1]*g[2*(ig_base+ir)+1];
          alp[2*(ia_base+ir)+1] += sh[1]*g[2*(ig_base+ir)+0] + sh[0]*g[2*(ig_base+ir)+1];
          /* derivatives */
          gdsht[0] = g[2*(ig_base+ir)+0]*dsht[0] - g[2*(ig_base+ir)+1]*dsht[1];
          gdsht[1] = g[2*(ig_base+ir)+0]*dsht[1] + g[2*(ig_base+ir)+1]*dsht[0];
          gdshp[0] = g[2*(ig_base+ir)+0]*dshp[0] - g[2*(ig_base+ir)+1]*dshp[1];
          gdshp[1] = g[2*(ig_base+ir)+0]*dshp[1] + g[2*(ig_base+ir)+1]*dshp[0];
          dalp[0][2*(ia_base+ir)+0] += x2*gdsht[0] + x3*gdshp[0];
          dalp[0][2*(ia_base+ir)+1] += x2*gdsht[1] + x3*gdshp[1];
          dalp[1][2*(ia_base+ir)+0] += y2*gdsht[0] + y3*gdshp[0];
          dalp[1][2*(ia_base+ir)+1] += y2*gdsht[1] + y3*gdshp[1];
          dalp[2][2*(ia_base+ir)+0] += z2*gdsht[0];
          dalp[2][2*(ia_base+ir)+1] += z2*gdsht[1];
        } /* --- loop of ir --- */
      } 
#if SUMSKIP
      if (dgmax[l1*lmax+l2]>SUMSKIP_THRESHOLD) 
#endif
      {
        for (ir=0; ir<ngrid; ir++) {
          /* derivatives */
          dgsh[0] = dg[2*(ig_base+ir)+0]*sh[0] - dg[2*(ig_base+ir)+1]*sh[1];
          dgsh[1] = dg[2*(ig_base+ir)+0]*sh[1] + dg[2*(ig_base+ir)+1]*sh[0];
          dalp[0][2*(ia_base+ir)+0] += x1*dgsh[0];
          dalp[0][2*(ia_base+ir)+1] += x1*dgsh[1];
          dalp[1][2*(ia_base+ir)+0] += y1*dgsh[0];
          dalp[1][2*(ia_base+ir)+1] += y1*dgsh[1];
          dalp[2][2*(ia_base+ir)+0] += z1*dgsh[0];
          dalp[2][2*(ia_base+ir)+1] += z1*dgsh[1];
        } /* --- loop of ir --- */
      } 
    } /* --- loop of i ---*/
#if SUMSKIP
    for (j1=0; j1<jmax; j1++) {
      ia_base  = j1*ngrid;
      for (ir=0; ir<ngrid; ir++) {
        ia = ia_base+ir;
        if (fabs(alp[2*ia+0])>amax[j1]) amax[j1] = fabs(alp[2*ia+0]);
        if (fabs(alp[2*ia+1])>amax[j1]) amax[j1] = fabs(alp[2*ia+1]);
        if (fabs(dalp[0][2*ia+0])>damax[0][j1]) damax[0][j1] = fabs(dalp[0][2*ia+0]);
        if (fabs(dalp[0][2*ia+1])>damax[0][j1]) damax[0][j1] = fabs(dalp[0][2*ia+1]);
        if (fabs(dalp[1][2*ia+0])>damax[1][j1]) damax[1][j1] = fabs(dalp[1][2*ia+0]);
        if (fabs(dalp[1][2*ia+1])>damax[1][j1]) damax[1][j1] = fabs(dalp[1][2*ia+1]);
        if (fabs(dalp[2][2*ia+0])>damax[2][j1]) damax[2][j1] = fabs(dalp[2][2*ia+0]);
        if (fabs(dalp[2][2*ia+1])>damax[2][j1]) damax[2][j1] = fabs(dalp[2][2*ia+1]);
      } /* loop of ir */
    } /* loop of j1 */
#endif
#else /* GAUNTTABLE */
    for (j1=0; j1<jmax; j1++) {
      l1 = j2l[j1];
      m1 = j2m[j1];
      ia_base = j1*ngrid;
      for (j2=0; j2<jmax; j2++) {
        l2 = j2l[j2];
        m2 = j2m[j2];
#if SUMSKIP
        if (gmax[l1*lmax+l2]>SUMSKIP_THRESHOLD) {
#endif
          gnt = ERI_Gaunt(l, m, l1, m1, l2, m2);
          fpg = gnt*4.0*PI;
          ERI_Spherical_Harmonics(l2, m2, theta, phi, sh, dsht, dshp);
          phase(l-l1+l2, sh);
          phase(l-l1+l2, dsht);
          phase(l-l1+l2, dshp); 
          sh[0]   = fpg*sh[0];
          sh[1]   = fpg*sh[1];
          dsht[0] = fpg*dsht[0];
          dsht[1] = fpg*dsht[1];
          dshp[0] = fpg*dshp[0];
          dshp[1] = fpg*dshp[1];
          ig_base  = (l1*lmax+l2)*ngrid;
          for (ir=0; ir<ngrid; ir++) {
            ia = ia_base+ir;
            ig = ig_base+ir;
            alp[2*ia+0] += sh[0]*g[2*ig+0] - sh[1]*g[2*ig+1];
            alp[2*ia+1] += sh[1]*g[2*ig+0] + sh[0]*g[2*ig+1];
            gdsht[0] = g[2*ig+0]*dsht[0] - g[2*ig+1]*dsht[1];
            gdsht[1] = g[2*ig+0]*dsht[1] + g[2*ig+1]*dsht[0];
            gdshp[0] = g[2*ig+0]*dshp[0] - g[2*ig+1]*dshp[1];
            gdshp[1] = g[2*ig+0]*dshp[1] + g[2*ig+1]*dshp[0];
            dalp[0][2*ia+0] += x2*gdsht[0] + x3*gdshp[0];
            dalp[0][2*ia+1] += x2*gdsht[1] + x3*gdshp[1];
            dalp[1][2*ia+0] += y2*gdsht[0] + y3*gdshp[0];
            dalp[1][2*ia+1] += y2*gdsht[1] + y3*gdshp[1];
            dalp[2][2*ia+0] += z2*gdsht[0];
            dalp[2][2*ia+1] += z2*gdsht[1];
          } /* --- loop of ir --- */
#if SUMSKIP
        }
        if (dgmax[l1*lmax+l2]>SUMSKIP_THRESHOLD) {
#endif
          for (ir=0; ir<ngrid; ir++) {
            ia = ia_base+ir;
            ig = ig_base+ir;
            dgsh[0]  = dg[2*ig+0]*sh[0] - dg[2*ig+1]*sh[1];
            dgsh[1]  = dg[2*ig+0]*sh[1] + dg[2*ig+1]*sh[0];
            dalp[0][2*ia+0] += x1*dgsh[0];
            dalp[0][2*ia+1] += x1*dgsh[1];
            dalp[1][2*ia+0] += y1*dgsh[0];
            dalp[1][2*ia+1] += y1*dgsh[1];
            dalp[2][2*ia+0] += z1*dgsh[0];
            dalp[2][2*ia+1] += z1*dgsh[1];
          } /* --- loop of ir --- */
#if SUMSKIP
        }
#endif
      } /* --- loop of j2 ---*/
#if SUMSKIP
      for (ir=0; ir<ngrid; ir++) {
        ia = ia_base+ir;
        if (fabs(alp[2*ia+0])>amax[j1]) amax[j1] = fabs(alp[2*ia+0]);
        if (fabs(alp[2*ia+1])>amax[j1]) amax[j1] = fabs(alp[2*ia+1]);
        if (fabs(dalp[0][2*ia+0])>damax[0][j1]) damax[0][j1] = fabs(dalp[0][2*ia+0]);
        if (fabs(dalp[0][2*ia+1])>damax[0][j1]) damax[0][j1] = fabs(dalp[0][2*ia+1]);
        if (fabs(dalp[1][2*ia+0])>damax[1][j1]) damax[1][j1] = fabs(dalp[1][2*ia+0]);
        if (fabs(dalp[1][2*ia+1])>damax[1][j1]) damax[1][j1] = fabs(dalp[1][2*ia+1]);
        if (fabs(dalp[2][2*ia+0])>damax[2][j1]) damax[2][j1] = fabs(dalp[2][2*ia+0]);
        if (fabs(dalp[2][2*ia+1])>damax[2][j1]) damax[2][j1] = fabs(dalp[2][2*ia+1]);
      } /* loop of ir */
#endif
    } /*--- loop of j1 ---*/
#endif /* GAUNTTABLE */
  }
}






/*----------------------------------------------------------------------
  ERI_LL_Overlap
  
  Product of two orbital
----------------------------------------------------------------------*/
void ERI_LL_Overlap(
  ERI_t        *ptr,
  double       *p,   /* (OUT) P term */
  const double *a1,  /* (IN)  Alpha term of the orbtial #1 */
  const double *a2   /* (IN)  Alpha term of the orbital #2 */
)
{
  int j, j1, j2, ir, ia1, ia2, ip;
  int ia1_base, ia2_base, ip_base;
  double gnt, a1a2[2];
#if GAUNTTABLE
  int i, n;
#else
  int l, m, l1, m1, l2, m2;
#endif
#if SUMSKIP
  const double *amax1, *amax2;
  double *pmax;
#endif
 
  int        ngrid = ERI_ngrid(ptr);
  int        lmax  = ERI_lmax(ptr);
  int        jmax  = lmax*lmax;
  const int *j2l   = ptr->j2l;
  const int *j2m   = ptr->j2m;
  const ERI_Gaunt_Table_t *gtbl = ptr->gtbl;

  STEPTRACE("ERI_LL_Overlap: in\n");

#if SUMSKIP
  amax1 = &a1[2*ngrid*jmax]; 
  amax2 = &a2[2*ngrid*jmax]; 
  pmax  = &p[2*ngrid*jmax]; 
#endif

  /* zero rest */
  for (j=0; j<jmax*ngrid*2; j++) { p[j] = 0.0; }
#if SUMSKIP
  for (j=0; j<jmax; j++) { pmax[j] = 0.0; }
#endif

#if GAUNTTABLE
  for (j=0; j<jmax; j++) {
    n = ERI_Gaunt_Table_Num_Items(gtbl, j);
    ip_base = j*ngrid;
    for (i=0; i<n; i++) {
      gnt = ERI_Gaunt_Table_Coeff(gtbl, j, i);
      ERI_Gaunt_Table_j1j2(gtbl, j, i, &j1, &j2);
      ia1_base = j1*ngrid;
      ia2_base = j2*ngrid;
#if SUMSKIP
//      if (amax1[j1]>SUMSKIP_THRESHOLD
//          && amax2[j2]>SUMSKIP_THRESHOLD) {
      if (amax1[j1]*amax2[j2]>SUMSKIP_THRESHOLD) {
#endif
        for (ir=0; ir<ngrid; ir++) {  
          ia1 = (ia1_base+ir)*2; 
          ia2 = (ia2_base+ir)*2; 
          ip = (ip_base+ir)*2;
          a1a2[0] = a1[ia1+0]*a2[ia2+0] - a1[ia1+1]*a2[ia2+1];
          a1a2[1] = a1[ia1+0]*a2[ia2+1] + a1[ia1+1]*a2[ia2+0];
          p[ip+0] += gnt*a1a2[0];
          p[ip+1] += gnt*a1a2[1];
        } /*--- loop of ir ---*/
#if SUMSKIP
      }
#endif
    } /*--- loop of i ---*/
#if SUMSKIP
    for (ir=0; ir<ngrid; ir++) {  
      ip = (ip_base+ir)*2;
      if (fabs(p[ip+0])>pmax[j]) pmax[j] = fabs(p[ip+0]);     
      if (fabs(p[ip+1])>pmax[j]) pmax[j] = fabs(p[ip+1]);      
    }
#endif
  } /*--- loop of j ---*/
#else /* GAUNTTABLE */
  for (j=0; j<jmax; j++) {
    l = j2l[j];
    m = j2m[j];
    ip_base = j*ngrid;
    for (j1=0; j1<jmax; j1++) {
      l1 = j2l[j1];
      m1 = j2m[j1];
      for (j2=0; j2<jmax; j2++) {
        l2 = j2l[j2];
        m2 = j2m[j2];
#if SUMSKIP
//        if (amax1[j1]>SUMSKIP_THRESHOLD
//          && amax2[j2]>SUMSKIP_THRESHOLD) {
        if (amax1[j1]*amax[j2]>SUMSKIP_THRESHOLD) {
#endif
          gnt = ERI_Gaunt(l, m, l1, m1, l2, m2);
          ia1_base = j1*ngrid;
          ia2_base = j2*ngrid;
          for (ir=0; ir<ngrid; ir++) {  
            ia1 = (ia1_base+ir)*2; 
            ia2 = (ia2_base+ir)*2; 
            ip = (ip_base+ir)*2;
            a1a2[0] = a1[ia1+0]*a2[ia2+0]-a1[ia1+1]*a2[ia2+1];
            a1a2[1] = a1[ia1+0]*a2[ia2+1]+a1[ia1+1]*a2[ia2+0];
            p[ip+0] += gnt*a1a2[0];
            p[ip+1] += gnt*a1a2[1];
          } /*--- loop of ir ---*/  
#if SUMSKIP
        }
#endif
      } /*--- loop of j2 ---*/
    } /*--- loop of j1 ---*/
#if SUMSKIP
    for (ir=0; ir<ngrid; ir++) {  
      ip = (ip_base+ir)*2;
      if (fabs(p[ip+0])>pmax[j]) pmax[j] = fabs(p[ip+0]);     
      if (fabs(p[ip+1])>pmax[j]) pmax[j] = fabs(p[ip+1]);      
    }
#endif
  } /*--- loop of j ---*/
#endif /* GAUNTTABLE */

  STEPTRACE("ERI_Overlap: out\n");
}



/*--------------------------------------------------------------------*/
void ERI_LL_Overlap_d(
  ERI_t        *ptr,
  double       *p,
  double       *dp[3],
  const double *a1,
  const double *da1[3],
  const double *a2,
  const double *da2[3],
  double        x
)
{
  int i, j, j1, j2, ir,  ia1, ia2, ip, ixyz;
  int ia1_base, ia2_base, ip_base;
  int l, m;
  double gnt, a1a2[2], da1a2[3][2], a1da2[3][2];
#if GAUNTTABLE
  int n;
#else
  int l1, m1, l2, m2;
#endif

#if SUMSKIP
  double *pmax, *dpmax[3];
  const double *amax1, *damax1[3], *amax2, *damax2[3];
#endif 
  
  int        ngrid = ERI_ngrid(ptr);
  int        lmax  = ERI_lmax(ptr);
  int        jmax  = lmax*lmax;
  const int *j2l   = ptr->j2l;
  const int *j2m   = ptr->j2m;
  const ERI_Gaunt_Table_t *gtbl = ptr->gtbl;

#if SUMSKIP
  pmax      = &p[2*ngrid*jmax];
  amax1     = &a1[2*ngrid*jmax];
  amax2     = &a2[2*ngrid*jmax];
  for (i=0; i<3; i++) {
    dpmax[i]  = &(dp[i][2*ngrid*jmax]);
    damax1[i] = &(da1[i][2*ngrid*jmax]);
    damax2[i] = &(da2[i][2*ngrid*jmax]);
  }
#endif
  STEPTRACE("MCP_P_d: begin\n");

  /* zero reset */
  for (j=0; j<jmax*ngrid*2; j++) {
    p[j] = 0.0;
    dp[0][j] = 0.0;
    dp[1][j] = 0.0;
    dp[2][j] = 0.0;
  }
#if SUMSKIP
  for (j=0; j<jmax; j++) {
    pmax[j]  = 0.0;
    dpmax[0][j]  = 0.0;
    dpmax[1][j]  = 0.0;
    dpmax[2][j]  = 0.0;
  }
#endif 
  STEPTRACE("MCP_P_d: step 1\n");


  for (j=0; j<jmax; j++) {
    l = j2l[j];
    m = j2m[j];
    ip_base = j*ngrid;
#if GAUNTTABLE  
    n = ERI_Gaunt_Table_Num_Items(gtbl, j);
    for (i=0; i<n; i++) {
      gnt = ERI_Gaunt_Table_Coeff(gtbl, j, i);
      ERI_Gaunt_Table_j1j2(gtbl, j, i, &j1, &j2);
      {
#else /* GAUNTTABLE */
    for (j1=0; j1<jmax; j1++) {
      l1 = j2l[j1];
      m1 = j2m[j1];
      for (j2=0; j2<jmax; j2++) {
        l2 = j2l[j1];
        m2 = j2m[j1];
        gnt = ERI_Gaunt(l, m, l1, m1, l2, m2);
#endif /*GAUNTTABLE*/
        ia1_base = j1*ngrid;
        ia2_base = j2*ngrid;
#if SUMSKIP
//        if (amax1[j1]>SUMSKIP_THRESHOLD
//            && amax2[j2]>SUMSKIP_THRESHOLD) {
        if (amax1[j1]*amax2[j2]>SUMSKIP_THRESHOLD) {
          for (ir=0; ir<ngrid; ir++) {  
            ia1 = ia1_base+ir; 
            ia2 = ia2_base+ir; 
            ip  = ip_base+ir;
            a1a2[0] = a1[2*ia1+0]*a2[2*ia2+0] - a1[2*ia1+1]*a2[2*ia2+1];
            a1a2[1] = a1[2*ia1+0]*a2[2*ia2+1] + a1[2*ia1+1]*a2[2*ia2+0];
            p[2*ip+0] += gnt*a1a2[0];
            p[2*ip+1] += gnt*a1a2[1];
          } /* loop of ir */
        }
//        if (amax2[j2]>SUMSKIP_THRESHOLD) {
//          for (ixyz=0; ixyz<3; ixyz++) {
//            if (damax1[ixyz][j1]>SUMSKIP_THRESHOLD) {
        for (ixyz=0; ixyz<3; ixyz++) {
          if (amax2[j2]*damax1[ixyz][j1]>SUMSKIP_THRESHOLD) {
            for (ir=0; ir<ngrid; ir++) {  
              ia1 = ia1_base+ir; 
              ia2 = ia2_base+ir; 
              ip  = ip_base+ir;
              da1a2[ixyz][0] = da1[ixyz][2*ia1+0]*a2[2*ia2+0] - da1[ixyz][2*ia1+1]*a2[2*ia2+1];
              da1a2[ixyz][1] = da1[ixyz][2*ia1+0]*a2[2*ia2+1] + da1[ixyz][2*ia1+1]*a2[2*ia2+0];
              dp[ixyz][2*ip+0] += gnt*(1.0-x)*da1a2[ixyz][0];
              dp[ixyz][2*ip+1] += gnt*(1.0-x)*da1a2[ixyz][1];
            } /* loop of ir */
          }
        } /* loop of ixyz */

//        if (amax1[j1]>SUMSKIP_THRESHOLD) {
//          for (ixyz=0; ixyz<3; ixyz++) {
//            if (damax2[ixyz][j2]>SUMSKIP_THRESHOLD) {
        for (ixyz=0; ixyz<3; ixyz++) {
          if (amax1[j1]*damax2[ixyz][j2]>SUMSKIP_THRESHOLD) {
            for (ir=0; ir<ngrid; ir++) {  
              ia1 = ia1_base+ir; 
              ia2 = ia2_base+ir; 
              ip  = ip_base+ir;
              a1da2[ixyz][0] = a1[2*ia1+0]*da2[ixyz][2*ia2+0] - a1[2*ia1+1]*da2[ixyz][2*ia2+1];
              a1da2[ixyz][1] = a1[2*ia1+0]*da2[ixyz][2*ia2+1] + a1[2*ia1+1]*da2[ixyz][2*ia2+0];
              dp[ixyz][2*ip+0] += -gnt*x*a1da2[ixyz][0];
              dp[ixyz][2*ip+1] += -gnt*x*a1da2[ixyz][1];
            } /* loop of ir */
          }
        } /* loop of ixyz */
#else /* SUMSKIP */
        for (ir=0; ir<ngrid; ir++) {  
          ia1 = (ia1_base+ir)*2; 
          ia2 = (ia2_base+ir)*2; 
          ip = (ip_base+ir)*2;
              
          a1a2[0] = a1[ia1+0]*a2[ia2+0]-a1[ia1+1]*a2[ia2+1];
          a1a2[1] = a1[ia1+0]*a2[ia2+1]+a1[ia1+1]*a2[ia2+0];
          p[ip+0] += gnt*a1a2[0];
          p[ip+1] += gnt*a1a2[1];
              
          da1a2[0][0] = da1[0][ia1+0]*a2[ia2+0]-da1[0][ia1+1]*a2[ia2+1];
          da1a2[0][1] = da1[0][ia1+0]*a2[ia2+1]+da1[0][ia1+1]*a2[ia2+0];
          da1a2[1][0] = da1[1][ia1+0]*a2[ia2+0]-da1[1][ia1+1]*a2[ia2+1];
          da1a2[1][1] = da1[1][ia1+0]*a2[ia2+1]+da1[1][ia1+1]*a2[ia2+0];
          da1a2[2][0] = da1[2][ia1+0]*a2[ia2+0]-da1[2][ia1+1]*a2[ia2+1];
          da1a2[2][1] = da1[2][ia1+0]*a2[ia2+1]+da1[2][ia1+1]*a2[ia2+0];
              
          a1da2[0][0] = a1[ia1+0]*da2[0][ia2+0]-a1[ia1+1]*da2[0][ia2+1];
          a1da2[0][1] = a1[ia1+0]*da2[0][ia2+1]+a1[ia1+1]*da2[0][ia2+0];
          a1da2[1][0] = a1[ia1+0]*da2[1][ia2+0]-a1[ia1+1]*da2[1][ia2+1];
          a1da2[1][1] = a1[ia1+0]*da2[1][ia2+1]+a1[ia1+1]*da2[1][ia2+0];
          a1da2[2][0] = a1[ia1+0]*da2[2][ia2+0]-a1[ia1+1]*da2[2][ia2+1];
          a1da2[2][1] = a1[ia1+0]*da2[2][ia2+1]+a1[ia1+1]*da2[2][ia2+0];
 
          dp[0][ip+0] += gnt*((1.0-x)*da1a2[0][0] - x*a1da2[0][0]);
          dp[0][ip+1] += gnt*((1.0-x)*da1a2[0][1] - x*a1da2[0][1]);
          dp[1][ip+0] += gnt*((1.0-x)*da1a2[1][0] - x*a1da2[1][0]);
          dp[1][ip+1] += gnt*((1.0-x)*da1a2[1][1] - x*a1da2[1][1]);
          dp[2][ip+0] += gnt*((1.0-x)*da1a2[2][0] - x*a1da2[2][0]);
          dp[2][ip+1] += gnt*((1.0-x)*da1a2[2][1] - x*a1da2[2][1]);
        } /*--- loop of ir ---*/
#endif /* SUMSKIP */
#if GAUNTTABLE 
      }
    } /*--- loop of i ---*/
#else
      } /* loop of j2 */
    } /* loop of j1 */
#endif 
#if SUMSKIP
    for (ir=0; ir<ngrid; ir++) {  
      ip = ip_base+ir;
      if (fabs(p[2*ip+0])>pmax[j]) pmax[j] = fabs(p[2*ip+0]);     
      if (fabs(p[2*ip+1])>pmax[j]) pmax[j] = fabs(p[2*ip+1]);      
      for (i=0; i<3; i++) {
        if (fabs(dp[i][2*ip+0])>dpmax[i][j]) dpmax[i][j] = fabs(dp[i][2*ip+0]);     
        if (fabs(dp[i][2*ip+1])>dpmax[i][j]) dpmax[i][j] = fabs(dp[i][2*ip+1]);      
      } /* loop of i */
    } /* loop of ir */
#endif
  } /*--- loop of j ---*/
}




void ERI_Transform_Overlap(
  ERI_t        *ptr,
  double       *F,
  const double *P
)
{
  int j, ip, l;
#if SUMSKIP
  int ik;
  double *fmax;
  const double *pmax;
#endif
  
  int ngrid = ERI_ngrid(ptr);
  int lmax  = ERI_lmax(ptr);
  int lmax0 = ERI_lmax_gl(ptr);
  int jmax  = lmax*lmax;
  int jmax0 = lmax0*lmax0;
  ERI_SBT_t *sbt = ptr->sbt;

  STEPTRACE("ERI_Transform_Overlap: in");

#if SUMSKIP
  pmax = &P[2*ngrid*jmax];
  fmax = &F[2*ngrid*jmax0]; 
  for (j=0; j<jmax0; j++) { fmax[j] = 0.0; }
#endif
  
  for (j=0; j<jmax0; j++) {
#if SUMSKIP
    if (pmax[j]>SUMSKIP_THRESHOLD) 
#endif
    {
      l=(int)sqrt((double)j);
      ip = j*ngrid;
      ERI_SBT_Transform(sbt, &F[2*ip], &P[2*ip], l, ERI_SBT_FORWARD);
#if SUMSKIP
      for (ik=0; ik<ngrid; ik++) {
        if (fabs(F[2*(ip+ik)+0])>fmax[j]) fmax[j] = fabs(F[2*(ip+ik)+0]);
        if (fabs(F[2*(ip+ik)+1])>fmax[j]) fmax[j] = fabs(F[2*(ip+ik)+1]);
      }
#endif
    }
  } /* loop of j */
  
  STEPTRACE("ERI_Transofrm_Overlap: out");
}




/*----------------------------------------------------------------------
  ERI_GL_Interpolate
----------------------------------------------------------------------*/
void ERI_GL_Interpolate(
  ERI_t        *ptr,
  double       *glF, /* (IN) Overlap matrix */
  const double *F    /* (IN) Overlap matrix */
)
{
  int j, ik, ig;
  double k, y[2];
#if SUMSKIP
  double *glfmax;
  const double *fmax;
#endif
  ERI_CSpline_t *cs;
 
  int           ngrid = ERI_ngrid(ptr);
  int           ngl   = ERI_ngl(ptr);
  int           lmax0 = ERI_lmax_gl(ptr);
  int           jmax0 = lmax0*lmax0;
  const double *kmesh = ERI_Mesh_Array_k(ptr);
  const double *glx   = ERI_Mesh_Array_glx(ptr);

#if SUMSKIP
  glfmax = &glF[ngl*jmax0*2];
  fmax   = &F[ngrid*jmax0*2];
  for (j=0; j<jmax0; j++) { glfmax[j] = 0.0; }
#endif

  for (j=0; j<jmax0; j++) {
#if SUMSKIP
    if (fmax[j]>SUMSKIP_THRESHOLD) 
#endif
    {
      cs = ERI_CSpline_Init_Complex(kmesh, (const double*)&F[j*ngrid*2], ngrid);
      for (ik=0; ik<ngl; ik++) {
        k = glx[ik];
        ig = (j*ngl+ik)*2;
        ERI_CSpline_Eval_Complex(y, k, cs);
        glF[ig+0] = y[0];
        glF[ig+1] = y[1];
#if SUMSKIP
        if (fabs(glF[ig+0])>glfmax[j]) glfmax[j]=fabs(glF[ig+0]);
        if (fabs(glF[ig+1])>glfmax[j]) glfmax[j]=fabs(glF[ig+1]);
#endif
      } /* loop of ik */
      ERI_CSpline_Free(cs);
    }
  } /* loop of j */
}



/*----------------------------------------------------------------------
  ERI_Integral_GL

  If you need the derivatives, you should call ERI_I4_d.
----------------------------------------------------------------------*/
void ERI_Integral_GL(
  ERI_t        *ptr,
  double        I4[2], /* (OUT) */
  const double *F1,    /* (IN) Overlap matrix */
  const double *F2,    /* (IN) */
  double        R,     /* (IN) Displacement of two expansion centers */
  double        theta,      
  double        phi,
  double        omega,  /* (IN) screening parameter */
  int           lmax1
)
{
  int i, j, j1, j2, ik;
  int gl_ibase1, gl_ibase2;
  int l, m, l1, l2;
#if GAUNTTABLE
  int n;
#else
  int m1, m2;
#endif
  static double tmp_sb[100];
  static double tmp_dsb[100];

  double sh[2], dsht[2], dshp[2];
  double k;
  double sum[2], gnt;
  double FG[2];
  double int1[2];
#if SUMSKIP
  const double *fmax1, *fmax2;
#endif

  int          ngl   = ERI_ngl(ptr);
  int          lmax  = ERI_lmax(ptr);
  int          lmax0 = ERI_lmax_gl(ptr);
  int          jmax0 = lmax0*lmax0;
  int          jmax1 = lmax1*lmax1;
  const double *glx  = ptr->glq_x;
  const double *glw  = ptr->glq_w;
  double       *glj  = ptr->glq_j;
  const int    *j2l  = ptr->j2l;
  const int    *j2m  = ptr->j2m;
  const ERI_Gaunt_Table_t *gtbl = ptr->gtbl;

  double scr, scr_a;

#if SUMSKIP
  fmax1 = &F1[2*ngl*jmax0];
  fmax2 = &F2[2*ngl*jmax0];
#endif

  /* spherical Bessel function */
  if (omega<0.0) {
    /* no screening */
    for (ik=0; ik<ngl; ik++) {
      k = glx[ik];
      ERI_Spherical_Bessel(k*R, lmax, tmp_sb, tmp_dsb);
      for (l=0; l<lmax; l++) {
        i = l*ngl + ik;
        glj[i] = glw[ik]*exp(k)*tmp_sb[l]*32.0*PI;
      }
    }
  } else {
    /* screening */
    scr_a = 0.25/omega/omega;
    for (ik=0; ik<ngl; ik++) {
      k = glx[ik];
      ERI_Spherical_Bessel(k*R, lmax, tmp_sb, tmp_dsb);
      scr = 1.0-exp(-k*k*scr_a);
      for (l=0; l<lmax; l++) {
        i = l*ngl + ik;
        glj[i] = glw[ik]*exp(k)*tmp_sb[l]*32.0*PI*scr;
      }
    }
  }


  I4[0] = 0.0;
  I4[1] = 0.0;
 
  //for (j=0; j<jmax0; j++) {
  for (j=0; j<jmax1; j++) {
    l = j2l[j];
    m = j2m[j];
    /* isbk_base = l*ngrid; */
    ERI_Spherical_Harmonics(l, m, theta, phi, sh, dsht, dshp);
    int1[0] = 0.0;
    int1[1] = 0.0;
#if GAUNTTABLE
    n = ERI_Gaunt_Table_Num_Items(gtbl, j);
    for (i=0; i<n; i++) {
      gnt = ERI_Gaunt_Table_Coeff(gtbl, j, i);
      ERI_Gaunt_Table_j1j2(gtbl, j, i, &j1, &j2);
      //if (j1<jmax0 && j2<jmax0) {
      if (j1<jmax1 && j2<jmax1) {
#else /* GAUNTTABLE */
    //for (j1=0; j1<jmax0; j1++) {
    for (j1=0; j1<jmax1; j1++) {
      l1 = j2l[j1];
      m1 = j2m[j1];
      //for (j2=0; j2<jmax0; j2++) {
      for (j2=0; j2<jmax1; j2++) {
        l2 = j2l[j2];
        m2 = j2m[j2];
        gnt = ERI_Gaunt(l, m, l1, m1, l2, m2);
#endif /*GAUNTTABLE */
#if SUMSKIP
//        if (fmax1[j1]>SUMSKIP_THRESHOLD
//            && fmax2[j2]>SUMSKIP_THRESHOLD) {
        if (fmax1[j1]*fmax2[j2]>SUMSKIP_THRESHOLD) {
#endif
          l1 = j2l[j1];
          l2 = j2l[j2];
          sum[0] = 0.0;
          sum[1] = 0.0;
          gl_ibase1 = j1*ngl;
          gl_ibase2 = j2*ngl;
          for (ik=0; ik<ngl; ik++) {
            FG[0] = F1[(gl_ibase1+ik)*2+0]*F2[(gl_ibase2+ik)*2+0]
                    -F1[(gl_ibase1+ik)*2+1]*F2[(gl_ibase2+ik)*2+1];
            FG[1] = F1[(gl_ibase1+ik)*2+0]*F2[(gl_ibase2+ik)*2+1]
                    +F1[(gl_ibase1+ik)*2+1]*F2[(gl_ibase2+ik)*2+0];
            sum[0] += glj[l*ngl+ik]*FG[0];
            sum[1] += glj[l*ngl+ik]*FG[1];
          } 
          phase(l2+l-l1, sum);
          int1[0] += sum[0]*gnt; 
          int1[1] += sum[1]*gnt; 
#if SUMSKIP
        }
#endif
#if GAUNTTABLE
      } 
    } /* loop of j1 */
#else
      } /* loop of j2 */
    } /* loop of j1 */
#endif 
    I4[0] += int1[0]*sh[0] - int1[1]*sh[1];
    I4[1] += int1[1]*sh[0] + int1[0]*sh[1];
  } /* loop of j */


}



/*----------------------------------------------------------------------
  ERI_Integral_GL_d
----------------------------------------------------------------------*/
void ERI_Integral_GL_d(
  ERI_t        *ptr,
  double        I4[2],  /* (OUT) */
  double        dI4[4][3][2],
  const double *F1,     /* (IN) Overlap matrix */
  const double *F2,     /* (IN) */
  const double *dF1[3],    /* (IN) Overlap matrix */
  const double *dF2[3],     /* (IN) */
  double        R,      /* (IN) Displacement of two expansion centers */
  double        theta,      
  double        phi,
  double        cx12,
  double        cx34,
  double        delta,
  double        omega,  /* (IN) screening parameter */
  int           lmax1
)
{
  int i, j, j1, j2, ik, ixyz;
  int gl_ibase1, gl_ibase2;
  int l, m, l1, l2;
#if GAUNTTABLE
  int n;
#else
  int m1, m2;
#endif
  static double tmp_sb[100];
  static double tmp_dsb[100];

  double sh[2], dsht[2], dshp[2];
  double k, gnt;
  double sum[2], dsum[2];
  double FG[2], dFG[3][2], FdG[3][2];
  double int1[2], int2[2], int3[3][2], int4[3][2];
  double q1[2], q2[2], q3[2], q4[2], q5[3][2], q6[2], q7[2];

  double ct, st, cp, sp;

  double x1, y1, z1, x2, y2, z2, x3, y3, Rd;

#if SUMSKIP
  const double *fmax1, *fmax2;
  const double *dfmax1[3], *dfmax2[3];
#endif

  int          ngl   = ERI_ngl(ptr);
  int          lmax  = ERI_lmax(ptr);
  int          lmax0 = ERI_lmax_gl(ptr);
  int          jmax0 = lmax0*lmax0;
  int          jmax1 = lmax1*lmax1;
  const double *glx  = ptr->glq_x;
  const double *glw  = ptr->glq_w;
  double       *glj  = ptr->glq_j;
  double       *gldj = ptr->glq_dj;
  const int    *j2l  = ptr->j2l;
  const int    *j2m  = ptr->j2m;
  const ERI_Gaunt_Table_t *gtbl = ptr->gtbl;

  double scr, scr_a;

#if SUMSKIP
  fmax1 = &F1[2*ngl*jmax0];
  fmax2 = &F2[2*ngl*jmax0];
  dfmax1[0] = &dF1[0][2*ngl*jmax0];
  dfmax1[1] = &dF1[1][2*ngl*jmax0];
  dfmax1[2] = &dF1[2][2*ngl*jmax0];
  dfmax2[0] = &dF2[0][2*ngl*jmax0];
  dfmax2[1] = &dF2[1][2*ngl*jmax0];
  dfmax2[2] = &dF2[2][2*ngl*jmax0];
#endif

  Rd = sqrt(R*R+delta*delta*exp(-R*R/delta/delta));
  ct = cos(theta);
  st = sin(theta);
  cp = cos(phi);
  sp = sin(phi);

  x1 = st*cp;    y1 = st*sp;    z1 = ct;
  x2 = ct*cp/Rd; y2 = ct*sp/Rd; z2 = -st/Rd;
  x3 = -sp/Rd/st; y3 = cp/Rd/st;

  if (fabs(theta)<1e-10) {
    x3 = 0.0;
    y3 = 0.0;
  }
  
  /* spherical Bessel function */
  if (omega<0.0) {
    /* no screening */
    for (ik=0; ik<ngl; ik++) {
      k = glx[ik];
      ERI_Spherical_Bessel(k*R, lmax, tmp_sb, tmp_dsb);
      for (l=0; l<lmax; l++) {
        i = l*ngl + ik;
        glj[i]  = glw[ik]*exp(k)*tmp_sb[l]*32.0*PI;
        gldj[i] = glw[ik]*exp(k)*tmp_dsb[l]*k*32.0*PI;
      }
    }
  } else {
    /* screening */
    scr_a = 0.25/omega/omega;
    for (ik=0; ik<ngl; ik++) {
      k = glx[ik];
      ERI_Spherical_Bessel(k*R, lmax, tmp_sb, tmp_dsb);
      scr = 1.0-exp(-k*k*scr_a);
      for (l=0; l<lmax; l++) {
        i = l*ngl + ik;
        glj[i]  = glw[ik]*exp(k)*tmp_sb[l]*32.0*PI*scr;
        gldj[i] = glw[ik]*exp(k)*tmp_dsb[l]*k*32.0*PI*scr;
      }
    }
  }


  I4[0] = 0.0;
  I4[1] = 0.0;
  for (i=0; i<4; i++) {
    for (ixyz=0; ixyz<3; ixyz++) {
      dI4[i][ixyz][0] = 0.0;
      dI4[i][ixyz][1] = 0.0;
    }
  }
 
  for (j=0; j<jmax1; j++) {
    l = j2l[j];
    m = j2m[j];
    /* isbk_base = l*ngrid; */
    ERI_Spherical_Harmonics(l, m, theta, phi, sh, dsht, dshp);
    int1[0] = 0.0;
    int1[1] = 0.0;
    int2[0] = 0.0;
    int2[1] = 0.0;
    for (ixyz=0; ixyz<3; ixyz++) {
      int3[ixyz][0] = 0.0;
      int3[ixyz][1] = 0.0;
      int4[ixyz][0] = 0.0;
      int4[ixyz][1] = 0.0;
    }
#if GAUNTTABLE
    n = ERI_Gaunt_Table_Num_Items(gtbl, j);
    for (i=0; i<n; i++) {
      gnt = ERI_Gaunt_Table_Coeff(gtbl, j, i);
      ERI_Gaunt_Table_j1j2(gtbl, j, i, &j1, &j2);
      if (j1<jmax1 && j2<jmax1) {
#else /* GAUNTTABLE */
    for (j1=0; j1<jmax1; j1++) {
      l1 = j2l[j1];
      m1 = j2m[j1];
      for (j2=0; j2<jmax1; j2++) {
        l2 = j2l[j2];
        m2 = j2m[j2];
        gnt = ERI_Gaunt(l, m, l1, m1, l2, m2);
#endif /*GAUNTTABLE */
        l1 = j2l[j1];
        l2 = j2l[j2];
        gl_ibase1 = j1*ngl;
        gl_ibase2 = j2*ngl;
#if SUMSKIP
//        if (fmax1[j1]>SUMSKIP_THRESHOLD
//            && fmax2[j2]>SUMSKIP_THRESHOLD) 
        if (fmax1[j1]*fmax2[j2]>SUMSKIP_THRESHOLD) 
#endif
        {
          sum[0] = 0.0;
          sum[1] = 0.0;
          dsum[0] = 0.0;
          dsum[1] = 0.0;
          for (ik=0; ik<ngl; ik++) {
            FG[0] = F1[(gl_ibase1+ik)*2+0]*F2[(gl_ibase2+ik)*2+0]
                    -F1[(gl_ibase1+ik)*2+1]*F2[(gl_ibase2+ik)*2+1];
            FG[1] = F1[(gl_ibase1+ik)*2+0]*F2[(gl_ibase2+ik)*2+1]
                    +F1[(gl_ibase1+ik)*2+1]*F2[(gl_ibase2+ik)*2+0];
              
            sum[0] += glj[l*ngl+ik]*FG[0];
            sum[1] += glj[l*ngl+ik]*FG[1];
            
            dsum[0] += gldj[l*ngl+ik]*FG[0];
            dsum[1] += gldj[l*ngl+ik]*FG[1];
          }
          phase(l2+l-l1, sum);
          phase(l2+l-l1, dsum);

          int1[0] += sum[0]*gnt; 
          int1[1] += sum[1]*gnt; 

          int2[0] += dsum[0]*gnt; 
          int2[1] += dsum[1]*gnt; 
        }
        for (ixyz=0; ixyz<3; ixyz++) {
#if SUMSKIP
//          if (dfmax1[ixyz][j1]>SUMSKIP_THRESHOLD
//              && fmax2[j2]>SUMSKIP_THRESHOLD) 
          if (dfmax1[ixyz][j1]*fmax2[j2]>SUMSKIP_THRESHOLD) 
#endif
          {
            dsum[0] = 0.0;
            dsum[1] = 0.0;
            for (ik=0; ik<ngl; ik++) {
              dFG[ixyz][0] = dF1[ixyz][(gl_ibase1+ik)*2+0]*F2[(gl_ibase2+ik)*2+0]
                            -dF1[ixyz][(gl_ibase1+ik)*2+1]*F2[(gl_ibase2+ik)*2+1];
              dFG[ixyz][1] = dF1[ixyz][(gl_ibase1+ik)*2+0]*F2[(gl_ibase2+ik)*2+1]
                            +dF1[ixyz][(gl_ibase1+ik)*2+1]*F2[(gl_ibase2+ik)*2+0];
           
              dsum[0] += glj[l*ngl+ik]*dFG[ixyz][0];
              dsum[1] += glj[l*ngl+ik]*dFG[ixyz][1];
            }
            phase(l2+l-l1, dsum);
            int3[ixyz][0] += dsum[0]*gnt; 
            int3[ixyz][1] += dsum[1]*gnt; 
          }
#if SUMSKIP
//          if (fmax1[j1]>SUMSKIP_THRESHOLD
//              && dfmax2[ixyz][j2]>SUMSKIP_THRESHOLD) 
          if (fmax1[j1]*dfmax2[ixyz][j2]>SUMSKIP_THRESHOLD) 
#endif
          {
            dsum[0] = 0.0;
            dsum[1] = 0.0;
            for (ik=0; ik<ngl; ik++) {
              FdG[ixyz][0] = F1[(gl_ibase1+ik)*2+0]*dF2[ixyz][(gl_ibase2+ik)*2+0]
                            -F1[(gl_ibase1+ik)*2+1]*dF2[ixyz][(gl_ibase2+ik)*2+1];
              FdG[ixyz][1] = F1[(gl_ibase1+ik)*2+0]*dF2[ixyz][(gl_ibase2+ik)*2+1]
                            +F1[(gl_ibase1+ik)*2+1]*dF2[ixyz][(gl_ibase2+ik)*2+0];
  
              dsum[0] += glj[l*ngl+ik]*FdG[ixyz][0];
              dsum[1] += glj[l*ngl+ik]*FdG[ixyz][1];
            } 
            phase(l2+l-l1, dsum);
            int4[ixyz][0] += dsum[0]*gnt; 
            int4[ixyz][1] += dsum[1]*gnt; 
          }
        } /* loop of ixyz */
#if GAUNTTABLE
      } 
    } /* loop of j1 */
#else
      } /* loop of j2 */
    } /* loop of j1 */
#endif 
 
    q1[0] = int1[0]*sh[0] - int1[1]*sh[1];
    q1[1] = int1[1]*sh[0] + int1[0]*sh[1];

    I4[0] += q1[0];
    I4[1] += q1[1];

    q2[0] = int2[0]*sh[0] - int2[1]*sh[1];
    q2[1] = int2[1]*sh[0] + int2[0]*sh[1];

    q3[0] = int1[0]*dsht[0] - int1[1]*dsht[1];
    q3[1] = int1[1]*dsht[0] + int1[0]*dsht[1];
    q4[0] = int1[0]*dshp[0] - int1[1]*dshp[1];
    q4[1] = int1[1]*dshp[0] + int1[0]*dshp[1];
   
    q5[0][0] = x1*q2[0] + x2*q3[0] + x3*q4[0]; 
    q5[0][1] = x1*q2[1] + x2*q3[1] + x3*q4[1]; 
    q5[1][0] = y1*q2[0] + y2*q3[0] + y3*q4[0]; 
    q5[1][1] = y1*q2[1] + y2*q3[1] + y3*q4[1]; 
    q5[2][0] = z1*q2[0] + z2*q3[0];
    q5[2][1] = z1*q2[1] + z2*q3[1];

    for (ixyz=0; ixyz<3; ixyz++) {
      q6[0] = int3[ixyz][0]*sh[0] - int3[ixyz][1]*sh[1];
      q6[1] = int3[ixyz][1]*sh[0] + int3[ixyz][0]*sh[1];
      q7[0] = int4[ixyz][0]*sh[0] - int4[ixyz][1]*sh[1];
      q7[1] = int4[ixyz][1]*sh[0] + int4[ixyz][0]*sh[1];
     
      dI4[0][ixyz][0] +=      -cx12*q5[ixyz][0] + q6[0];
      dI4[0][ixyz][1] +=      -cx12*q5[ixyz][1] + q6[1];
      dI4[1][ixyz][0] += (cx12-1.0)*q5[ixyz][0] - q6[0];
      dI4[1][ixyz][1] += (cx12-1.0)*q5[ixyz][1] - q6[1];
      dI4[2][ixyz][0] +=       cx34*q5[ixyz][0] + q7[0];
      dI4[2][ixyz][1] +=       cx34*q5[ixyz][1] + q7[1];
      dI4[3][ixyz][0] += (1.0-cx34)*q5[ixyz][0] - q7[0];
      dI4[3][ixyz][1] += (1.0-cx34)*q5[ixyz][1] - q7[1];
    } 
  } /* loop of j */
}



/* EOF */
